# Copyright: (c) 2019, Jordan Borean (@jborean93) <jborean93@gmail.com> and other contributors
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import errno
import ntpath
import os
import os.path
import shutil
import stat
import sys

from smbclient._io import SMBFileTransaction, SMBRawIO, query_info, set_info
from smbclient._os import SMBDirEntry
from smbclient._os import copyfile as smbclient_copyfile
from smbclient._os import (
    is_remote_path,
    makedirs,
    open_file,
    readlink,
    remove,
    rmdir,
    scandir,
)
from smbclient._os import stat as smbclient_stat
from smbclient._os import symlink
from smbclient.path import isdir, islink, samefile
from smbprotocol import MAX_PAYLOAD_SIZE
from smbprotocol.file_info import FileAttributes, FileBasicInformation
from smbprotocol.open import CreateOptions, FilePipePrinterAccessMask
from smbprotocol.structure import DateTimeField


def _basename(path):
    """
    Return the base name of pathname path.
    This is the second element of the pair returned by passing path to the function split().
    Note that the result of this function is different from the Unix basename program;
    where basename for '/foo/bar/' returns 'bar', the basename() function returns an empty string ('').

    :param path: The path to take the base of.
    :return: The basename of the path.
    """
    if is_remote_path(path):
        return ntpath.basename(path)
    else:
        return os.path.basename(path)


def _join_local_or_remote_path(path, *paths):
    """
    Return the joined paths regardless of whether they are local or remote.
    Currently, this only makes the decision based on the base path, and does not raise a ValueError if the additional
    paths do not match.
    :param path: The base path.
    :param paths: The additional paths to append.
    :return: Joined path.
    """
    if is_remote_path(path):
        return ntpath.join(path, *paths)
    else:
        return os.path.join(path, *paths)


def copy(src, dst, follow_symlinks=True, **kwargs):
    """
    Copies the file src to the file or directory dst. If dst specified a directory, the file will be copied into dst
    using the base filename from src. Returns the path to the newly created file.

    If follow_symlinks is 'False', and src is a symbolic link, dst will be created as a symbolic link. If
    follow_symlinks is 'True' and src is a symbolic link, dst will be a copy of the file src refers to.

    copy() copies the file data and file's permission mode which on Windows is only the read-only flag. All other
    bits are ignored on a copy. Other metadata like file's creation and modification times, are not preserved.

    :param src: The path to the source file to copy, if not an absolute path then this is relative to the current
        directory.
    :param dst: The path to the destination to copy the file t.
    :param follow_symlinks: Whether to copy the symlink target of source or the symlink itself.
    :param kwargs: Common arguments used to build the SMB Session for any UNC paths.
    :return The path to the destination filename.
    """
    return _copy(src, dst, follow_symlinks, copymode, **kwargs)


def copy2(src, dst, follow_symlinks=True, **kwargs):
    """
    Identical to copy() except that copy2() also attempts to preserve the file metadata.

    copy2() uses copystat() to copy the file metadata. Please see copystat() for more information about how and what
    metadata it copies to the dst file.

    :param src: The path to the source file to copy, if not an absolute path then this is relative to the current
        directory.
    :param dst: The path to the destination to copy the file t.
    :param follow_symlinks: Whether to copy the symlink target of source or the symlink itself.
    :param kwargs: Common arguments used to build the SMB Session for any UNC paths.
    :return The path to the destination filename.
    """
    return _copy(src, dst, follow_symlinks, copystat, **kwargs)


def copyfile(src, dst, follow_symlinks=True, **kwargs):
    """
    Copy the contents (no metadata) of the file names src to a file named dst and return dst in the most efficient way
    possible. If the src and dst reside on the same UNC share then a more efficient remote side copy is used. If the
    src and dst reside on the same local path then the proper shutil.copyfile() method is called, otherwise the content
    is copied using copyfileobj() in chunks.

    dst must be the complete target file name; look at copy() for a copy that accepts a target directory path. If src
    dst specify the same file, ValueError is raised.

    The destination location must be writable; otherwise, an OSError exception will be raised. If dst already exists,
    it will be replaced. Special files such as character or block devices and pipes cannot be copied with this
    function.

    If follow_symlinks is 'False' and src is a symbolic link, a new symbolic link will be created instead of copying
    the file src points to. This will fail if the symbolic link at src is a different root as dst.

    :param src: The src filepath to copy.
    :param dst: The destination filepath to copy to.
    :param follow_symlinks: Whether to copy the symlink target of source or the symlink itself.
    :param kwargs: Common arguments used to build the SMB Session for any UNC paths.
    :return: The dst path.
    """
    norm_src = ntpath.normpath(src)
    if is_remote_path(norm_src):
        src_root = ntpath.splitdrive(norm_src)[0]
        islink_func = islink
        readlink_func = readlink
        symlink_func = symlink
        src_open = open_file
        src_kwargs = kwargs
    else:
        src_root = None
        islink_func = os.path.islink
        readlink_func = os.readlink
        symlink_func = os.symlink
        src_open = open
        src_kwargs = {}

    norm_dst = ntpath.normpath(dst)
    if is_remote_path(norm_dst):
        dst_root = ntpath.splitdrive(norm_dst)[0]
        dst_open = open_file
        dst_kwargs = kwargs
    else:
        dst_root = None
        dst_open = open
        dst_kwargs = {}

    if not follow_symlinks and islink_func(src, **src_kwargs):
        if src_root != dst_root:
            raise ValueError("Cannot copy a symlink on different roots.")

        symlink_func(readlink_func(src), dst, **src_kwargs)
        return dst

    if src_root is None and dst_root is None:
        # The files are local and follow_symlinks is True, rely on the builtin copyfile mechanism.
        shutil.copyfile(src, dst)
        return dst

    if src_root == dst_root:
        # The files are located on the same share and follow_symlinks is True, rely on smbclient.copyfile for an
        # efficient server side copy.
        try:
            is_same = samefile(src, dst, **kwargs)
        except OSError as err:
            if err.errno == errno.ENOENT:
                is_same = False
            else:
                raise

        if is_same:
            raise shutil.Error(f"'{src}' and '{dst}' are the same file, cannot copy")

        smbclient_copyfile(src, dst, **kwargs)
        return dst

    # Finally we are copying across different roots so we just chunk the data using copyfileobj
    with src_open(src, mode="rb", **src_kwargs) as src_fd, dst_open(dst, mode="wb", **dst_kwargs) as dst_fd:
        copyfileobj(src_fd, dst_fd, MAX_PAYLOAD_SIZE)

    return dst


# Because smbclient's open_file returns a file object this should just work.
copyfileobj = shutil.copyfileobj


def copymode(src, dst, follow_symlinks=True, **kwargs):
    """
    Copy the permission bits from src to dst. The file contents, owner, and group are unaffected. Due to the
    limitations of Windows, this function only sets/unsets the dst's FILE_ATTRIBUTE_READ_ONLY flag based on what src's
    attribute is set to.

    If follow_symlinks is 'False', and both src and dst are symbolic links, copymode() will attempt to modify the mode
    of dst itself (rather than the file it points to).

    This function supports src and dst being either a local or UNC path. A relative path will be resolved based on the
    current working directory.

    :param src: The src file or directory to copy the read only flag from.
    :param dst: The dst file or directory to copy the read only flag to.
    :param follow_symlinks: Whether to copy the read only flag on the symlink or the target of the symlink.
    :param kwargs: Common arguments used to build the SMB Session for any UNC paths.
    """
    src_mode = stat.S_IMODE(_get_file_stat(src, follow_symlinks, **kwargs).st_mode)

    norm_dst = ntpath.normpath(dst)
    if is_remote_path(norm_dst):
        read_only = not (src_mode & stat.S_IWRITE == stat.S_IWRITE and src_mode & stat.S_IREAD == stat.S_IREAD)
        _set_file_basic_info(dst, follow_symlinks, read_only=read_only, **kwargs)
    else:
        os.chmod(dst, src_mode, follow_symlinks=follow_symlinks)


def copystat(src, dst, follow_symlinks=True, **kwargs):
    """
    Copy the read only attribute, last access time, and last modification time from src to dst. The file contents,
    owner, and group are unaffected.

    If follow_symlinks is 'False' and src and dst both refer to symbolic links, copystat() will operate on the
    symbolic links themselves rather than the files the symbolic links refer to.

    :param src: The src file or directory to copy the read only flag from.
    :param dst: The dst file or directory to copy the read only flag to.
    :param follow_symlinks: Whether to copy the read only flag on the symlink or the target of the symlink.
    :param kwargs: Common arguments used to build the SMB Session for any UNC paths.
    """
    src_stat = _get_file_stat(src, follow_symlinks, **kwargs)
    src_mode = stat.S_IMODE(src_stat.st_mode)

    atime_ns = src_stat.st_atime_ns
    mtime_ns = src_stat.st_mtime_ns

    norm_dst = ntpath.normpath(dst)
    if is_remote_path(norm_dst):
        read_only = not (src_mode & stat.S_IWRITE == stat.S_IWRITE and src_mode & stat.S_IREAD == stat.S_IREAD)
        _set_file_basic_info(dst, follow_symlinks, read_only=read_only, atime_ns=atime_ns, mtime_ns=mtime_ns, **kwargs)
    else:
        os.chmod(dst, src_mode, follow_symlinks=follow_symlinks)
        os.utime(dst, ns=(atime_ns, mtime_ns), follow_symlinks=follow_symlinks)


def copytree(
    src,
    dst,
    symlinks=False,
    ignore=None,
    copy_function=copy2,
    ignore_dangling_symlinks=False,
    dirs_exist_ok=False,
    **kwargs,
):
    """
    Recursively copy an entire directory tree rooted at src to a directory named dst and return the destination
    directory. dirs_exist_ok dictates whether to raise an exception in case dst or any missing parent directory
    exists.

    Permissions and times of directories are copied with copystat(), individual files are copied using copy2().

    If symlinks is 'True', symbolic links in the source tree are represented as symbolic links in the new tree and the
    metadata of the original links will be copied as far as the platform allow; if 'False' or omitted, the contents and
    metadata of the linked files are copied to the new tree.

    When symlinks is 'False', if the file pointed by the symlink doesn't exist, an exception will be added in the list
    of errors raises in an 'Error' exception at the end of the copy process. You can set the optional
    ignore_dangling_symlinks flag to 'True' if you want to silence this exception.

    if ignore is given, it must be a callable that will receive as its arguments the directory being visited by
    copytree(), and a list of its contents, as returned by smbclient.listdir(). Since copytee() is called recursively,
    the ignore callable will be called once for each directory that is copied. The callable must return a sequence of
    directory and file names relative to the current directory (i.e. a subset of the items in its second argument);
    these names will then be ignored in the copy process.

    If exception(s) occur, an shutil.Error is raised with a list of reasons.

    If copy_function is given, it must be a callable that will be used to copy each file. It will be called with the
    source path and the destination path as arguments. By default copy() is used, but any function that supports the
    same signature (like copy()) can be used.

    :param src: The source directory to copy.
    :param dst: The destination directory to copy to.
    :param symlinks: Whether to attempt to copy a symlink from the source tree to the dest tree, if False the symlink
        target's contents are copied instead as a normal file/dir.
    :param ignore: A callable in the form 'callable(src, names) -> ignored_named' that returns a list of file names to
        ignore based on the list passed in.
    :param copy_function: The copy function to use for copying files.
    :param ignore_dangling_symlinks: Ignore any broken symlinks, otherwise an Error is raised.
    :param dirs_exist_ok: Whether to fail if the dst directory exists or not.
    :param kwargs: Common arguments used to build the SMB Session for any UNC paths.
    :return: The dst path.
    """
    if is_remote_path(src):
        dir_entries = list(scandir(src, **kwargs))
    else:
        dir_entries = list(os.scandir(src))

    if is_remote_path(dst):
        makedirs(dst, exist_ok=dirs_exist_ok, **kwargs)
    else:
        os.makedirs(dst, exist_ok=dirs_exist_ok)

    ignored = []
    if ignore is not None:
        ignored = ignore(src, [e.name for e in dir_entries])

    errors = []
    for dir_entry in dir_entries:
        if dir_entry.name in ignored:
            continue

        src_path = _join_local_or_remote_path(src, dir_entry.name)
        dst_path = _join_local_or_remote_path(dst, dir_entry.name)

        try:
            if dir_entry.is_symlink():
                if not isinstance(dir_entry, SMBDirEntry):
                    raise AssertionError("copytree doesn't yet support symlinks for local to remote operations")

                link_target = readlink(src_path, **kwargs)
                if symlinks:
                    symlink(link_target, dst_path, **kwargs)
                    copystat(src_path, dst_path, follow_symlinks=False)
                    continue
                else:
                    # Manually override the dir_entry with a new one that is the link target and copy that below.
                    try:
                        dir_entry = SMBDirEntry.from_path(link_target, **kwargs)
                    except OSError as err:
                        if err.errno == errno.ENOENT and ignore_dangling_symlinks:
                            continue
                        raise

            if dir_entry.is_dir():
                copytree(
                    src_path,
                    dst_path,
                    symlinks,
                    ignore,
                    copy_function,
                    ignore_dangling_symlinks,
                    dirs_exist_ok,
                    **kwargs,
                )
            else:
                copy_function(src_path, dst_path, **kwargs)
        except shutil.Error as err:
            # From a recursive call of copytree().
            errors.extend(err.args[0])
        except OSError as err:
            # Main smbclient operations should raise an OSError exception (or one that inherits OSError).
            errors.append((src_path, dst_path, str(err)))
        except ValueError as err:
            # If the path is not supported or we are trying to symlink outside of the share boundary.
            errors.append((src_path, dst_path, str(err)))

    try:
        copystat(src, dst)
    except OSError as err:
        errors.append((src, dst, str(err)))

    if errors:
        raise shutil.Error(errors)
    return dst


def rmtree(path, ignore_errors=False, onerror=None, **kwargs):
    """
    Recursively delete a directory tree; path must point to a directory (but not a symbolic link to a directory).

    If ignore_errors is 'True', errors resulting from failed removals will be ignored; otherwise such errors are
    handled by calling a handler specified by onerror or, if that is omitted, they raise an exception.

    if onerror is provided, it must be a callable that accepts three parameters: function, path, excinfo.

    The first parameter, function, is the function which raised the exception. The second parameter, path, will be the
    path name passed to function. The third parameter, excinfo, will be the exception information returned by
    sys.exc_info(). Exceptions raised by onerror will not be caught.

    :param path: The path to remove.
    :param ignore_errors: Whether to ignore errors on failed removed or delegate the handling to onerror.
    :param onerror: The callback executed when an error on removal is raised.
    :param kwargs: Common arguments used to build the SMB Session.
    """
    if ignore_errors:

        def onerror(*args):
            pass

    elif onerror is None:

        def onerror(*args):
            raise

    if islink(path, **kwargs):
        try:
            raise OSError("Cannot call rmtree on a symbolic link")
        except OSError:
            onerror(islink, path, sys.exc_info())
            return

    scandir_gen = scandir(path, **kwargs)
    while True:
        try:
            dir_entry = next(scandir_gen)
        except StopIteration:
            break
        except OSError:
            onerror(scandir, path, sys.exc_info())
            continue

        # In case the entry is a directory symbolic link we need to remove the dir itself and not recurse down into
        # it with rmtree. Doing that would result in a symbolic link target having it's contents removed even if it's
        # outside the rmtree scope.
        if (
            dir_entry.is_symlink()
            and dir_entry.stat(follow_symlinks=False).st_file_attributes & FileAttributes.FILE_ATTRIBUTE_DIRECTORY
        ):
            try:
                rmdir(dir_entry.path, **kwargs)
            except OSError:
                onerror(rmdir, dir_entry.path, sys.exc_info())
        elif dir_entry.is_dir():
            rmtree(dir_entry.path, ignore_errors, onerror, **kwargs)
        else:
            try:
                remove(dir_entry.path, **kwargs)
            except OSError:
                onerror(remove, dir_entry.path, sys.exc_info())

    try:
        rmdir(path, **kwargs)
    except OSError:
        onerror(rmdir, path, sys.exc_info())


def _copy(src, dst, follow_symlinks, copy_meta_func, **kwargs):
    # Need to check if dst is a UNC path before checking if it's a dir in smbclient.path before checking to see if it's
    # a local directory. If either one is a dir, join the filename of src onto dst.
    if (is_remote_path(ntpath.normpath(dst)) and isdir(dst, **kwargs)) or os.path.isdir(dst):
        dst = _join_local_or_remote_path(dst, _basename(src))

    copyfile(src, dst, follow_symlinks=follow_symlinks)
    copy_meta_func(src, dst, follow_symlinks=follow_symlinks)
    return dst


def _get_file_stat(path, follow_symlinks=True, **kwargs):
    if path.startswith("//") or path.startswith("\\\\"):
        return smbclient_stat(path, follow_symlinks=follow_symlinks, **kwargs)
    else:
        # Source is a local path or accessible to the host, use the builtin os module to get the read only flag.
        if follow_symlinks:
            return os.stat(path)
        else:
            return os.lstat(path)


def _set_file_basic_info(file_path, follow_symlinks, read_only=None, atime_ns=None, mtime_ns=None, **kwargs):
    co = 0 if follow_symlinks else CreateOptions.FILE_OPEN_REPARSE_POINT

    # Need to check if we need to change the attributes if the read_only flag is set. We can't just blindly set the
    # read only attribute as that will remove any other attribute set on the file.
    new_attributes = 0
    if read_only is not None:
        file_fd = SMBRawIO(
            file_path,
            mode="r",
            share_access="rw",
            create_options=co,
            desired_access=FilePipePrinterAccessMask.FILE_READ_ATTRIBUTES,
            **kwargs,
        )
        with SMBFileTransaction(file_fd) as transaction:
            query_info(transaction, FileBasicInformation)

        existing_attributes = transaction.results[0]["file_attributes"].get_value()
        if read_only and existing_attributes & FileAttributes.FILE_ATTRIBUTE_READONLY == 0:
            new_attributes = existing_attributes | FileAttributes.FILE_ATTRIBUTE_READONLY
        elif not read_only and existing_attributes & FileAttributes.FILE_ATTRIBUTE_READONLY != 0:
            new_attributes = existing_attributes & ~FileAttributes.FILE_ATTRIBUTE_READONLY

            # Make sure at least 1 attribute is set (normal)
            if new_attributes == 0:
                new_attributes |= FileAttributes.FILE_ATTRIBUTE_NORMAL

    # Only set the attributes if there is actually a change to be made.
    if new_attributes or atime_ns is not None or mtime_ns is not None:
        file_fd = SMBRawIO(
            file_path,
            mode="r",
            share_access="rw",
            create_options=co,
            desired_access=FilePipePrinterAccessMask.FILE_WRITE_ATTRIBUTES,
            **kwargs,
        )

        with SMBFileTransaction(file_fd) as transaction:
            basic_info = FileBasicInformation()
            basic_info["file_attributes"] = new_attributes

            if atime_ns is not None:
                basic_info["last_access_time"] = int((atime_ns // 100) + DateTimeField.EPOCH_FILETIME)

            if mtime_ns is not None:
                basic_info["last_write_time"] = int((mtime_ns // 100) + DateTimeField.EPOCH_FILETIME)

            set_info(transaction, basic_info)



# Copyright: (c) 2019, Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

from collections import OrderedDict

from smbprotocol.header import NtStatus
from smbprotocol.structure import (
    BoolField,
    BytesField,
    DateTimeField,
    EnumField,
    FlagField,
    IntField,
    Structure,
    UuidField,
)


class CreateContextName:
    """
    [MS-SMB2] v53.0 2017-09-15

    2.2.13.2 SMB2_CREATE_CONTEXT Request Values
    Valid names for the name to set on a SMB2_CREATE_CONTEXT Request entry
    """

    SMB2_CREATE_EA_BUFFER = b"\x45\x78\x74\x41"

    # note: the structures for this are located in security_descriptor.py
    SMB2_CREATE_SD_BUFFER = b"\x53\x65\x63\x44"
    SMB2_CREATE_DURABLE_HANDLE_REQUEST = b"\x44\x48\x6e\x51"
    SMB2_CREATE_DURABLE_HANDLE_RECONNECT = b"\x44\x48\x6e\x43"
    SMB2_CREATE_ALLOCATION_SIZE = b"\x41\x6c\x53\x69"
    SMB2_CREATE_QUERY_MAXIMAL_ACCESS_REQUEST = b"\x4d\x78\x41\x63"
    SMB2_CREATE_TIMEWARP_TOKEN = b"\x54\x57\x72\x70"
    SMB2_CREATE_QUERY_ON_DISK_ID = b"\x51\x46\x69\x64"
    SMB2_CREATE_REQUEST_LEASE = b"\x52\x71\x4c\x73"
    SMB2_CREATE_REQUEST_LEASE_V2 = b"\x52\x71\x4c\x73"
    SMB2_CREATE_DURABLE_HANDLE_REQUEST_V2 = b"\x44\x48\x32\x51"
    SMB2_CREATE_DURABLE_HANDLE_RECONNECT_V2 = b"\x44\x48\x32\x43"
    SMB2_CREATE_APP_INSTANCE_ID = b"\x45\xBC\xA6\x6A\xEF\xA7\xF7\x4A" b"\x90\x08\xFA\x46\x2E\x14\x4D\x74"
    SMB2_CREATE_APP_INSTANCE_VERSION = b"\xB9\x82\xD0\xB7\x3B\x56\x07\x4F" b"\xA0\x7B\x52\x4A\x81\x16\xA0\x10"
    SVHDX_OPEN_DEVICE_CONTEXT = b"\x9C\xCB\xCF\x9E\x04\xC1\xE6\x43" b"\x98\x0E\x15\x8D\xA1\xF6\xEC\x83"

    @staticmethod
    def get_response_structure(name, size=None):
        """
        Returns the response structure for a know list of create context
        responses.

        :param name: The constant value above
        :param size: Specify the size of the context buffer, used to differenciate between REQUEST_LEASE and
            REQUEST_LEASE_V2.
        :return: The response structure or None if unknown
        """
        # Special handling for request lease here the header name has the same value, use the size to differenciate.
        if name == CreateContextName.SMB2_CREATE_REQUEST_LEASE:
            return {
                32: SMB2CreateResponseLease(),
                52: SMB2CreateResponseLeaseV2(),
            }.get(size, None)

        return {
            CreateContextName.SMB2_CREATE_DURABLE_HANDLE_REQUEST: SMB2CreateDurableHandleResponse(),
            CreateContextName.SMB2_CREATE_DURABLE_HANDLE_RECONNECT: SMB2CreateDurableHandleReconnect(),
            CreateContextName.SMB2_CREATE_QUERY_MAXIMAL_ACCESS_REQUEST: SMB2CreateQueryMaximalAccessResponse(),
            CreateContextName.SMB2_CREATE_QUERY_ON_DISK_ID: SMB2CreateQueryOnDiskIDResponse(),
            CreateContextName.SMB2_CREATE_DURABLE_HANDLE_REQUEST_V2: SMB2CreateDurableHandleResponseV2(),
            CreateContextName.SMB2_CREATE_DURABLE_HANDLE_RECONNECT_V2: SMB2CreateDurableHandleReconnectV2,
            CreateContextName.SMB2_CREATE_APP_INSTANCE_ID: SMB2CreateAppInstanceId(),
            CreateContextName.SMB2_CREATE_APP_INSTANCE_VERSION: SMB2CreateAppInstanceVersion(),
        }.get(name, None)


class EAFlags:
    """
    [MS-FSCC]

    2.4.15 FileFullEaInformation Flags
    Specifies the flag used when setting extended attributes.
    """

    NONE = 0x0000000
    FILE_NEED_EA = 0x00000080


class LeaseState:
    """
    [MS-SMB2]

    2.2.13.2.8 SMB2_CREATE_REQUEST_LEASE LeaseState
    The requested lease state, field is constructed with a combination of the
    following values.
    """

    SMB2_LEASE_NONE = 0x00
    SMB2_LEASE_READ_CACHING = 0x01
    SMB2_LEASE_HANDLE_CACHING = 0x02
    SMB2_LEASE_WRITE_CACHING = 0x04


class LeaseRequestFlags:
    """
    [MS-SMB2]

    2.2.13.2.10 SMB2_CREATE_REQUEST_LEASE_V2
    The flags to use on an SMB2CreateRequestLeaseV2 packet.
    """

    SMB2_LEASE_FLAG_PARENT_LEASE_KEY_SET = 0x00000004


class LeaseResponseFlags:
    """
    [MS-SMB2]

    2.2.14.2.10 SMB2_CREATE_RESPONSE_LEASE
    """

    SMB2_LEASE_FLAG_BREAK_IN_PROGRESS = 0x00000002
    SMB2_LEASE_FLAG_PARENT_LEASE_KEY_SET = 0x00000004  # V2 Response


class DurableHandleFlags:
    """
    [MS-SMB2]

    2.2.13.2.11 SMB2_CREATE_DURABLE_HANDLE_REQUEST_V2
    Flags used on an SMB2CreateDurableHandleRequestV2 packet.
    """

    SMB2_DHANDLE_FLAG_PERSISTENT = 0x00000002


class SVHDXOriginatorFlags:
    """
    [MS-RSVD] 2.2.4.12 SVHDX_OPEN_DEVICE_CONTEXT OriginatorFlags
    Used to indicate which component has originated or issued the operations.
    """

    SVHDX_ORIGINATOR_PVHDPARSER = 0x00000001
    SVHDX_ORIGINATOR_VHDMP = 0x00000004


class SMB2CreateContextRequest(Structure):
    """
    [MS-SMB2] v53.0 2017-09-15

    2.2.13.2 SMB2_CREATE_CONTEXT Request Values
    Structure used in the SMB2 CREATE Request and SMB2 CREATE Response to
    encode additional flags and attributes
    """

    def __init__(self):
        self.fields = OrderedDict(
            [
                ("next", IntField(size=4)),
                ("name_offset", IntField(size=2, default=16)),
                ("name_length", IntField(size=2, default=lambda s: len(s["buffer_name"]))),
                ("reserved", IntField(size=2)),
                ("data_offset", IntField(size=2, default=lambda s: self._buffer_data_offset(s))),
                ("data_length", IntField(size=4, default=lambda s: len(s["buffer_data"]))),
                ("buffer_name", BytesField(size=lambda s: s["name_length"].get_value())),
                (
                    "padding",
                    BytesField(size=lambda s: self._padding_size(s), default=lambda s: b"\x00" * self._padding_size(s)),
                ),
                ("buffer_data", BytesField(size=lambda s: s["data_length"].get_value())),
                # not actually a field but each list entry must start at the 8 byte
                # alignment
                (
                    "padding2",
                    BytesField(
                        size=lambda s: self._padding2_size(s), default=lambda s: b"\x00" * self._padding2_size(s)
                    ),
                ),
            ]
        )
        super().__init__()

    def _buffer_data_offset(self, structure):
        if structure["data_length"].get_value() == 0:
            return 0
        else:
            return structure["name_offset"].get_value() + len(structure["buffer_name"]) + len(structure["padding"])

    def _padding_size(self, structure):
        if structure["data_length"].get_value() == 0:
            return 0

        buffer_name_len = structure["name_length"].get_value()
        mod = buffer_name_len % 8
        return mod if mod == 0 else 8 - mod

    def _padding2_size(self, structure):
        data_length = len(structure["buffer_name"]) + len(structure["padding"]) + len(structure["buffer_data"])
        mod = data_length % 8
        return mod if mod == 0 else 8 - mod

    def get_context_data(self):
        """
        Get the buffer_data value of a context response and try to convert it
        to the relevant structure based on the buffer_name used. If it is an
        unknown structure then the raw bytes are returned.

        :return: relevant Structure of buffer_data or bytes if unknown name
        """
        buffer_name = self["buffer_name"].get_value()
        structure = CreateContextName.get_response_structure(buffer_name, size=self["data_length"].get_value())
        if structure:
            structure.unpack(self["buffer_data"].get_value())
            return structure
        else:
            # unknown structure, just return the raw bytes
            return self["buffer_data"].get_value()

    @staticmethod
    def pack_multiple(messages):
        """
        Converts a list of SMB2CreateContextRequest structures and packs them
        as a bytes object used when setting to the SMB2CreateRequest
        buffer_contexts field. This should be used as it would calculate the
        correct next field value for each context entry.

        :param messages: List of SMB2CreateContextRequest structures
        :return: bytes object that is set on the SMB2CreateRequest
            buffer_contexts field.
        """
        data = b""
        msg_count = len(messages)
        for i, msg in enumerate(messages):
            if not isinstance(msg, SMB2CreateContextRequest):
                buffer = msg
                buffer_name = getattr(msg, "NAME", None)
                if buffer_name is None:
                    raise ValueError(
                        "Invalid context message, must be either a SMB2CreateContextRequest or a "
                        "predefined structure object with NAME defined."
                    )
                msg = SMB2CreateContextRequest()
                msg["buffer_name"] = buffer_name
                msg["buffer_data"] = buffer

            if i == msg_count - 1:
                msg["next"] = 0
            else:
                # because the end padding2 val won't be populated if the entry
                # offset is 0, we set to 1 so the len calc is correct
                msg["next"] = 1
                msg["next"] = len(msg)

            data += msg.pack()
        return data


class SMB2CreateEABuffer(Structure):
    """
    [MS-SMB2] 2.2.13.2.1 SMB2_CREATE_EA_BUFFER
    [MS-FSCC] 2.4.15 FileFullEaInformation

    Used to apply extended attributes as part of creating a new file.
    """

    NAME = CreateContextName.SMB2_CREATE_EA_BUFFER

    def __init__(self):
        self.fields = OrderedDict(
            [
                # 0 if no more entries, otherwise offset after ea_value
                ("next_entry_offset", IntField(size=4)),
                ("flags", FlagField(size=1, flag_type=EAFlags)),
                ("ea_name_length", IntField(size=1, default=lambda s: len(s["ea_name"]) - 1)),  # minus \x00
                ("ea_value_length", IntField(size=2, default=lambda s: len(s["ea_value"]))),
                # ea_name is ASCII byte encoded and needs a null terminator '\x00'
                ("ea_name", BytesField(size=lambda s: s["ea_name_length"].get_value() + 1)),
                ("ea_value", BytesField(size=lambda s: s["ea_value_length"].get_value())),
                # not actually a field but each list entry must start at the 4 byte
                # alignment
                (
                    "padding",
                    BytesField(size=lambda s: self._padding_size(s), default=lambda s: b"\x00" * self._padding_size(s)),
                ),
            ]
        )
        super().__init__()

    def _padding_size(self, structure):
        if structure["next_entry_offset"].get_value() == 0:
            return 0

        data_length = len(structure["ea_name"]) + len(structure["ea_value"])
        mod = data_length % 4
        return mod if mod == 0 else 4 - mod

    @staticmethod
    def pack_multiple(messages):
        """
        Converts a list of SMB2CreateEABuffer structures and packs them as a
        bytes object used when setting to the SMB2CreateContextRequest
        buffer_data field. This should be used as it would calculate the
        correct next_entry_offset field value for each buffer entry.

        :param messages: List of SMB2CreateEABuffer structures
        :return: bytes object that is set on the SMB2CreateContextRequest
            buffer_data field.
        """
        data = b""
        msg_count = len(messages)
        for i, msg in enumerate(messages):
            if i == msg_count - 1:
                msg["next_entry_offset"] = 0
            else:
                # because the end padding val won't be populated if the entry
                # offset is 0, we set to 1 so the len calc is correct
                msg["next_entry_offset"] = 1
                msg["next_entry_offset"] = len(msg)
            data += msg.pack()

        return data


class SMB2CreateDurableHandleRequest(Structure):
    """
    [MS-SMB2] 2.2.13.2.3 SMB2_CREATE_DURABLE_HANDLE_REQUEST

    Used by the client to mark the open as a durable open.
    """

    NAME = CreateContextName.SMB2_CREATE_DURABLE_HANDLE_REQUEST

    def __init__(self):
        self.fields = OrderedDict([("durable_request", BytesField(size=16, default=b"\x00" * 16))])
        super().__init__()


class SMB2CreateDurableHandleResponse(Structure):
    """
    [MS-SMB2] 2.2.14.2.3 SMB2_CREATE_DURABLE_HANDLE_RESPONSE

    Sent by the server in response to an SMB2CreateDurableHandleRequest packet.
    """

    def __init__(self):
        self.fields = OrderedDict([("reserved", IntField(size=8))])
        super().__init__()


class SMB2CreateDurableHandleReconnect(Structure):
    """
    [MS-SMB2] 2.2.13.2.4 SMB2_CREATE_DURABLE_HANDLE_RECONNECT

    Used by the client when attempting to reestablish a durable open
    """

    NAME = CreateContextName.SMB2_CREATE_DURABLE_HANDLE_RECONNECT

    def __init__(self):
        self.fields = OrderedDict([("data", BytesField(size=16))])
        super().__init__()


class SMB2CreateQueryMaximalAccessRequest(Structure):
    """
    [MS-SMB2] 2.2.13.2.5 SMB2_CREATE_QUERY_MAXIMAL_ACCESS_REQUEST

    Used by the client to retrieve maximal access information as part of
    processing the open.
    """

    NAME = CreateContextName.SMB2_CREATE_QUERY_MAXIMAL_ACCESS_REQUEST

    def __init__(self):
        self.fields = OrderedDict([("timestamp", DateTimeField())])
        super().__init__()


class SMB2CreateQueryMaximalAccessResponse(Structure):
    """
    [MS-SMB2] 2.2.14.2.5 SMB2_CREATE_QUERY_MAXIMAL_ACCESS_RESPONSE

    Used by the server in response to an SMB2CreateQueryMaximalAccessRequest
    packet.
    """

    def __init__(self):
        self.fields = OrderedDict(
            [
                ("query_status", EnumField(size=4, enum_type=NtStatus, enum_strict=False)),
                # either FilePipePrinterAccessMask or DirectoryAccessMask
                ("maximal_access", IntField(size=4)),
            ]
        )
        super().__init__()


class SMB2CreateAllocationSize(Structure):
    """
    [MS-SMB2] 2.2.13.2.6 SMB2_CREATE_ALLOCATION_SIZE

    Used by the client to set the allocation size of a file that is being
    newly created or overwritten.
    """

    NAME = CreateContextName.SMB2_CREATE_ALLOCATION_SIZE

    def __init__(self):
        self.fields = OrderedDict([("allocation_size", IntField(size=8))])
        super().__init__()


class SMB2CreateTimewarpToken(Structure):
    """
    [MS-SMB2] 2.2.13.2.7 SMB2_CREATE_TIMEWARP_TOKEN

    Used by the client when requesting the server to open a version of the file
    at a previous point in time.
    """

    NAME = CreateContextName.SMB2_CREATE_TIMEWARP_TOKEN

    def __init__(self):
        self.fields = OrderedDict([("timestamp", DateTimeField())])
        super().__init__()


class SMB2CreateRequestLease(Structure):
    """
    [MS-SMB2] 2.2.13.2.8 SMB2_CREATE_REQUEST_LEASE

    Used by the cliet when requesting the server to return a lease.
    """

    NAME = CreateContextName.SMB2_CREATE_REQUEST_LEASE

    def __init__(self):
        self.fields = OrderedDict(
            [
                ("lease_key", BytesField(size=16)),
                ("lease_state", FlagField(size=4, flag_type=LeaseState)),
                ("lease_flags", IntField(size=4)),
                ("lease_duration", IntField(size=8)),
            ]
        )
        super().__init__()


class SMB2CreateResponseLease(Structure):
    """
    [MS-SMB2] 2.2.14.2.10 SMB2_CREATE_RESPONSE_LEASE

    Sent by the server in response to an SMB2CreateRequestLease
    """

    def __init__(self):
        self.fields = OrderedDict(
            [
                ("lease_key", BytesField(size=16)),
                ("lease_state", FlagField(size=4, flag_type=LeaseState)),
                ("lease_flags", FlagField(size=4, flag_type=LeaseResponseFlags)),
                ("lease_duration", IntField(size=8)),
            ]
        )
        super().__init__()


class SMB2CreateQueryOnDiskIDResponse(Structure):
    """
    [MS-SMB2] 2.2.14.2.9 SMB2_CREATE_QUERY_ON_DISK_ID

    Sent by the server in response to an SMB2CreateQueryOnDiskIDRequest packet.
    """

    NAME = CreateContextName.SMB2_CREATE_QUERY_ON_DISK_ID

    def __init__(self):
        self.fields = OrderedDict(
            [
                ("disk_file_id", IntField(size=8)),
                ("volume_id", IntField(size=8)),
                ("reserved", BytesField(size=16, default=b"\x00" * 16)),
            ]
        )
        super().__init__()


class SMB2CreateRequestLeaseV2(Structure):
    """
    [MS-SMB2] 2.2.13.2.10 SMB2_CREATE_REQUEST_LEASE_V2

    Used when the client is requesting the server to return a lease on a file
    or directory.
    Valid for the SMB 3.x family only
    """

    NAME = CreateContextName.SMB2_CREATE_REQUEST_LEASE_V2

    def __init__(self):
        self.fields = OrderedDict(
            [
                ("lease_key", BytesField(size=16)),
                ("lease_state", FlagField(size=4, flag_type=LeaseState)),
                ("lease_flags", FlagField(size=4, flag_type=LeaseRequestFlags)),
                ("lease_duration", IntField(size=8)),
                ("parent_lease_key", BytesField(size=16)),
                ("epoch", BytesField(size=2)),
                ("reserved", IntField(size=2)),
            ]
        )
        super().__init__()


class SMB2CreateResponseLeaseV2(Structure):
    """
    [MS-SMB2] 2.2.14.2.11 SMB2_CREATE_RESPONSE_LEASE_V2

    Sent by the server in response to an SMB2CreateRequestLeaseV2 packet.
    """

    def __init__(self):
        self.fields = OrderedDict(
            [
                ("lease_key", BytesField(size=16)),
                ("lease_state", FlagField(size=4, flag_type=LeaseState)),
                ("flags", FlagField(size=4, flag_type=LeaseResponseFlags)),
                ("lease_duration", IntField(size=8)),
                ("parent_lease_key", BytesField(size=16)),
                ("epoch", IntField(size=2)),
                ("reserved", IntField(size=2)),
            ]
        )
        super().__init__()


class SMB2CreateDurableHandleRequestV2(Structure):
    """
    [MS-SMB2] 2.2.13.2.11 SMB2_CREATE_DURABLE_HANDLE_REQUEST_V2

    Used by the client to request the server mark the open as durable or
    persistent.
    Valid for the SMB 3.x family only
    """

    NAME = CreateContextName.SMB2_CREATE_DURABLE_HANDLE_REQUEST_V2

    def __init__(self):
        self.fields = OrderedDict(
            [
                # timeout in milliseconds
                ("timeout", IntField(size=4)),
                ("flags", FlagField(size=4, flag_type=DurableHandleFlags)),
                ("reserved", IntField(size=8)),
                ("create_guid", UuidField(size=16)),
            ]
        )
        super().__init__()


class SMB2CreateDurableHandleReconnectV2(Structure):
    """
    [MS-SMB2] 2.2.13.2.12 SMB2_CREATE_DURABLE_HANDLE_RECONNECT_V2

    Used by the client when reestablishing a durable open.
    Valid for the SMB 3.x family only
    """

    NAME = CreateContextName.SMB2_CREATE_DURABLE_HANDLE_RECONNECT_V2

    def __init__(self):
        self.fields = OrderedDict(
            [
                ("file_id", BytesField(size=16)),
                ("create_guid", UuidField(size=16)),
                ("flags", FlagField(size=4, flag_type=DurableHandleFlags)),
            ]
        )
        super().__init__()


class SMB2CreateDurableHandleResponseV2(Structure):
    """
    [MS-SMB2] 2.2.14.2.12 SMB2_CREATE_DURABLE_HANDLE_RESPONSE_V2

    Sent by the server in response to an SMB2CreateDurableHandleRequestV2
    packet.
    """

    def __init__(self):
        self.fields = OrderedDict(
            [("timeout", IntField(size=4)), ("flags", FlagField(size=4, flag_type=DurableHandleFlags))]
        )
        super().__init__()


class SMB2CreateAppInstanceId(Structure):
    """
    [MS-SMB2] 2.2.13.2.13 SMB2_CREATE_APP_INSTANCE_ID

    Used by the client when supplying an identifier provided by an application.
    Valid for the SMB 3.x family and should also have an durable handle on the
    create request.
    """

    NAME = CreateContextName.SMB2_CREATE_APP_INSTANCE_ID

    def __init__(self):
        self.fields = OrderedDict(
            [
                ("structure_size", IntField(size=2, default=20)),
                ("reserved", IntField(size=2)),
                ("app_instance_id", BytesField(size=16)),
            ]
        )
        super().__init__()


class SMB2SVHDXOpenDeviceContextRequest(Structure):
    """
    [MS-SMB2] 2.2.13.2.14 SVHDX_OPEN_DEVICE_CONTEXT
    [MS-RSVD] 2.2.4.12 SVHDX_OPEN_DEVICE_CONTEXT

    Used to open the shared virtual disk file.
    """

    NAME = CreateContextName.SVHDX_OPEN_DEVICE_CONTEXT

    def __init__(self):
        self.fields = OrderedDict(
            [
                ("version", IntField(size=4, default=1)),
                ("has_initiator_id", BoolField(size=1, default=lambda s: len(s["initiator_host_name"]) > 0)),
                ("reserved", BytesField(size=3, default=b"\x00\x00\x00")),
                ("initiator_id", UuidField(size=16)),
                ("originator_flags", EnumField(size=4, enum_type=SVHDXOriginatorFlags)),
                ("open_request_id", IntField(size=8)),
                ("initiator_host_name_length", IntField(size=2, default=lambda s: len(s["initiator_host_name"]))),
                # utf-16-le encoded string
                ("initiator_host_name", BytesField(size=lambda s: s["initiator_host_name_length"].get_value())),
            ]
        )
        super().__init__()


class SMB2SVHDXOpenDeviceContextResponse(Structure):
    """
    [MS-SMB2] 2.2.14.2.14 SVHDX_OPEN_DEVICE_CONTEXT_RESPONSE
    [MS-RSVD] 2.2.4.31  SVHDX_OPEN_DEVICE_CONTEXT_RESPONSE

    The response packet sent by the server in response to an
    SMB2VHDXOpenDeviceContextRequest
    """

    def __init__(self):
        self.fields = OrderedDict(
            [
                ("version", IntField(size=4, default=1)),
                ("has_initiator_id", BoolField(size=1, default=lambda s: len(s["initiator_host_name"]) > 0)),
                ("reserved", BytesField(size=3, default=b"\x00\x00\x00")),
                ("initiator_id", UuidField(size=16)),
                ("flags", IntField(size=4)),
                ("originator_flags", EnumField(size=4, enum_type=SVHDXOriginatorFlags)),
                ("open_request_id", IntField(size=8)),
                ("initiator_host_name_length", IntField(size=2, default=lambda s: len(s["initiator_host_name"]))),
                # utf-16-le encoded string
                ("initiator_host_name", BytesField(size=lambda s: s["initiator_host_name_length"].get_value())),
            ]
        )
        super().__init__()


class SMB2SVHDXOpenDeviceContextV2Request(Structure):
    """
    [MS-SMB2] 2.2.13.2.14 SVHDX_OPEN_DEVICE_CONTEXT
    [MS-RSVD] 2.2.4.32 SVHDX_OPEN_DEVICE_CONTEXT_V2

    Used to open the shared virtual disk file on the RSVD Protocol version 2
    """

    NAME = CreateContextName.SVHDX_OPEN_DEVICE_CONTEXT

    def __init__(self):
        self.fields = OrderedDict(
            [
                ("version", IntField(size=4, default=2)),
                ("has_initiator_id", BoolField(size=1, default=lambda s: len(s["initiator_host_name"]) > 0)),
                ("reserved", BytesField(size=3, default=b"\x00\x00\x00")),
                ("initiator_id", UuidField(size=16)),
                ("originator_flags", EnumField(size=4, enum_type=SVHDXOriginatorFlags)),
                ("open_request_id", IntField(size=8)),
                ("initiator_host_name_length", IntField(size=2, default=lambda s: len(s["initiator_host_name"]))),
                # utf-16-le encoded string
                ("initiator_host_name", BytesField(size=lambda s: s["initiator_host_name_length"].get_value())),
                ("virtual_disk_properties_initialized", IntField(size=4)),
                ("server_service_version", IntField(size=4)),
                ("virtual_sector_size", IntField(size=4)),
                ("physical_sector_size", IntField(size=4)),
                ("virtual_size", IntField(size=8)),
            ]
        )
        super().__init__()


class SMB2SVHDXOpenDeviceContextV2Response(Structure):
    """
    [MS-SMB2] 2.2.14.2.14 SVHDX_OPEN_DEVICE_CONTEXT_RESPONSE
    [MS-RSVD] 2.2.4.32 SVHDX_OPEN_DEVICE_CONTEXT_V2_RESPONSE

    The response packet sent by the server in response to an
    SMB2VHDXOpenDeviceContextV2Request
    """

    def __init__(self):
        self.fields = OrderedDict(
            [
                ("version", IntField(size=4, default=2)),
                ("has_initiator_id", BoolField(size=1, default=lambda s: len(s["initiator_host_name"]) > 0)),
                ("reserved", BytesField(size=3, default=b"\x00\x00\x00")),
                ("initiator_id", UuidField(size=16)),
                ("flags", IntField(size=4)),
                ("originator_flags", EnumField(size=4, enum_type=SVHDXOriginatorFlags)),
                ("open_request_id", IntField(size=8)),
                ("initiator_host_name_length", IntField(size=2, default=lambda s: len(s["initiator_host_name"]))),
                # utf-16-le encoded string
                ("initiator_host_name", BytesField(size=lambda s: s["initiator_host_name_length"].get_value())),
                ("virtual_disk_properties_initialized", IntField(size=4)),
                ("server_service_version", IntField(size=4)),
                ("virtual_sector_size", IntField(size=4)),
                ("physical_sector_size", IntField(size=4)),
                ("virtual_size", IntField(size=8)),
            ]
        )
        super().__init__()


class SMB2CreateAppInstanceVersion(Structure):
    """
    [MS-SMB2] 2.2.13.2.15 SMB2_CREATE_APP_INSTANCE_VERSION

    Used when the client is supplying a version for the app instance identifier
    provided by an application.
    Valid for the SMB 3.1.1+ family
    """

    NAME = CreateContextName.SMB2_CREATE_APP_INSTANCE_VERSION

    def __init__(self):
        self.fields = OrderedDict(
            [
                ("structure_size", IntField(size=2, default=24)),
                ("reserved", IntField(size=2)),
                ("padding", IntField(size=4)),
                ("app_instance_version_high", IntField(size=8)),
                ("app_instance_version_low", IntField(size=8)),
            ]
        )
        super().__init__()
